local AceGUI = LibStub("AceGUI-3.0")
local LibST  = LibStub("ScrollingTable")
if (not AceGUI) or (not LibST) then return end

--------------------------
-- ScrollingTable		--
-- lib-st beta wrapper	--
-- farmbuyer			--
--------------------------

--[[
lib-st does not recycle the objects (called "ST" here) that it creates and
returns.  We therefore do not try to hold onto an ST when the widget is
being recycled.  This means that Constructor() does very little work, and
does not actually construct an ST.

OnAcquire cannot construct an ST either, because we don't yet have any
creation parameters from the user to feed to CreateST.  (Allowing such to
be passed along from AceGUI:Create() would require changes to core AceGUI
code, and I don't feel like trying to overcome that inertia.)

The upshot is that the widget returned from Create is broken and useless
until its CreateST member has been called.  This means that correct behavior
depends entirely on the user remembering to do so.

"The gods do not protect fools.  Fools are protected by more capable fools."
- Ringworld
]]

do
	local err = "Oops!  The AceGUI function you tried to call (%s) does not "
	            .. "make sense with lib-st and has not been implemented."
	local Type = "lib-st"
	local Version = 1

	--[[
	All of an ST's subframes are attached to its main frame, which we have in
	the st.frame link, and that's what AceGUI uses for all positioning.  Except
	that ST:SetDisplayCols creates its "head" row /above/ the main frame, and
	so the row of labels eats into whatever upper border space AceGUI calculates,
	often overlapping other elements.

	We get around this by replacing ST's main frame's SetPoint with a custom
	version that just moves everything down a few pixels to allow room for the
	head row.

	FIXME this may need to be a secure hook (ugh, would end up calling the real
	setpoint twice) rather than a replacement.
	]]
	local DEFAULT_OFFSET = 7
	local function ShiftingSetPoint(frame,anchor,other,otheranchor,xoff,yoff)
		local o = frame.obj.head_offset
		yoff = yoff or 0
		if anchor:sub(1,3) == "TOP" then
			yoff = yoff - o
		end
		return frame.realSetPoint(frame,anchor,other,otheranchor,xoff, (yoff or 0) - o)
	end
	local function ShiftingSetAllPoints(frame,other)
		ShiftingSetPoint(frame,"TOPLEFT",other,"TOPLEFT",0,0)
		ShiftingSetPoint(frame,"BOTTOMRIGHT",other,"BOTTOMRIGHT",0,0)
	end

	local function Oops(self)
		-- if you ever wanted an example of "brown paper bag" code, here it is
		local ds = debugstack(0)
		local func = ds:match("AceGUIWidget%-lib%-st%.lua:%d+:%s+in function `(%a+)'")
		error(err:format(func or "?"))
	end

	local function OnAcquire(self)
        -- Almost nothing can usefully be done here.
		self.head_offset = DEFAULT_OFFSET
	end

	local function OnRelease(self)
		self.st.frame:ClearAllPoints()
		self.st:Hide()
		self.st = nil
		-- XXX should also undo the frame hooks.  would most likely be wasted
		-- cycles.  if somebody actually wants to make an ST, include it inside
		-- an ace container, then hide the container and continue displaying
		-- the ST by other means, they can file a ticket
	end

	-- Purely for convenience.
	local function GetSTLibrary()
		return LibST
	end

	--[[
	STs don't use a "normal" SetWidth, if we define "normal" to be the
	behavior of the blizzard :SetWidth.  Column width is passed in during
	creation of the whole ST.  The SetWidth defined by an ST takes no
	arguments; "ReCalculateWidth" would be a more precise description of what
	it does.
	
	Parts of AceGUI look for a .width field because a widget's SetWidth sets
	such.  ST calculates a total width and dispatches it to its member frame...
	but doesn't store a local copy.  We need to bridge these differences.

	This widget wrapper does not make use of On{Width,Height}Set hooks, but
	the acegui widget base functions do.  Since we're not inheriting them,
	we may as well supply them.
	]]
	local function SetWidth(self)
		self.st:SetWidth()                    -- re-total the columns
		local w = self.st.frame:GetWidth()    -- fetch the answer back
		self.frame.width = w                  -- store it for acegui
		if self.OnWidthSet then
			self:OnWidthSet(w)
		end
	end

	-- Everything said about SetWidth applies here too.
	local function SetHeight(self)
		self.st:SetHeight()
		local h = self.st.frame:GetHeight()
		self.frame.height = h
		if self.OnHeightSet then
			self:OnHeightSet(h)
		end
	end

	-- Replacement wrapper, so that instead of
	--     st = ScrollingTable:CreateST( args )
	-- the user should be able to do
	--     st = AceGUI:Create("lib-st"):CreateST( args )
	-- instead.
	local function CreateST(self, ...)
		local st = LibST:CreateST(...)
		return self:WrapST(st)
	end

	--[[
	Users with an ST already constructed can drop it into a widget directly
	using this routine.  It must be safe to call this more than once with
	new widgets on the same ST.

	This is where most of the intelligence of the wrapper is located.  That
	is, if you can call my code "intelligent" with a straight face.  Lemme
	try again.

	Think of the widget wrapper as a brain.  When ALL THREE neurons manage
	to fire at the same time and produce a thought, this function represents
	the thought.  Sigh.
	]]
	local function WrapST(self, st)
		if not st.frame then
			error"lib-st instance has no '.frame' field... wtf did you pass to this function?"
		end
		if st.frame.obj and (st.frame.obj ~= self) then
			error"lib-st instance already has an '.obj' field from a different widget, cannot use with AceGUI!"
		end
		self.st = st
		if not st.head then
			error"lib-st instance has no '.head' field, must use either ScrollingTable:CreateST or this widget's CreatST first"
		end
		self.frame = st.frame   -- gutsy, but looks doable

		-- Possibly have already wrapped this ST in a previous widget, careful.
		if st.frame.obj ~= self then
			self.frame.realSetPoint = self.frame.SetPoint
			self.frame.SetPoint = ShiftingSetPoint
			self.frame.SetAllPoints = ShiftingSetAllPoints
		end

		-- This needs the .frame field.  This also creates .obj inside that
		-- field and calls a SetScript as well.
		AceGUI:RegisterAsWidget(self)

		return self
	end


	local function Constructor()
		local self = {}
		self.type = Type
		-- self.frame not set here, see WrapST

		-- Expected by AceGUI containers (and AceGUI users)
		self.OnAcquire = OnAcquire
		self.OnRelease = OnRelease
		self.SetWidth = SetWidth
		self.SetHeight = SetHeight

		-- Functions specific to this widget
		self.GetSTLibrary = GetSTLibrary
		self.CreateST = CreateST
		self.WrapST = WrapST

		-- Some AceGUI functions simply Won't Work in this context.  Name them
		-- here, and code calling them will get a somewhat informative error().
		for _,f in ipairs{
			'SetRelativeWidth', 'SetRelativeHeight',
			'SetFullWidth', 'SetFullHeight',
		} do self[f] = Oops end

		--AceGUI:RegisterAsWidget(self)  needs .frame
		return self
	end

	AceGUI:RegisterWidgetType(Type,Constructor,Version)
end

